/*
 *  linux/arch/arm/mach-pxa/xyz3.c
 *
 *  Support for the Intel HCDDBBVA0 Development Platform.
 *  (go figure how they came up with such name...)
 *
 *  Author:	Nicolas Pitre
 *  Created:	Nov 05, 2002
 *  Copyright:	MontaVista Software Inc.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/sysdev.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/bitops.h>
#include <linux/fb.h>

#include <asm/types.h>
#include <asm/setup.h>
#include <asm/memory.h>
#include <asm/mach-types.h>
#include <asm/hardware.h>
#include <asm/irq.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/irq.h>

#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>
#include <asm/arch/audio.h>
#include <asm/arch/pxafb.h>
#include <asm/arch/mmc.h>

#include "generic.h"

static void __init xyz3_init_irq(void)
{
	pxa_init_irq();
}

static struct resource smc91x_resources[] = {
	[0] = {
		.start	= (XYZ3_ETH_PHYS + 0x300),
		.end	= (XYZ3_ETH_PHYS + 0xfffff),
		.flags	= IORESOURCE_MEM,
	},
	[1] = {
		.start  = IRQ_GPIO(10),
		.end    = IRQ_GPIO(10),
		.flags	= IORESOURCE_IRQ,
	}
};

static struct platform_device smc91x_device = {
	.name		= "smc91x",
	.id		= 0,
	.num_resources	= ARRAY_SIZE(smc91x_resources),
	.resource	= smc91x_resources,
};


static struct platform_device xyz3_audio_device = {
	.name		= "pxa2xx-ac97",
	.id		= -1,
//	.dev		= { .platform_data = &xyz3_audio_ops },
};

static struct platform_device xyz3ostimer_device = {
	.name		= "xyz3-ostimer",
	.id		= -1,
};

static struct platform_device xyz3adc_device = {
	.name		= "xyz3-adc",
	.id		= -1,
	.dev	= { .parent = &xyz3ostimer_device.dev },
};

static struct platform_device xyz3pwm_device = {
	.name		= "xyz3-pwm",
	.id		= -1,
};

static struct platform_device xyz3gpio_device = {
	.name		= "xyz3-gpio",
	.id		= -1,
};

static struct platform_device xyz3doc_device = {
	.name		= "xyz3-doc",
	.id		= -1,
};

static void __init xyz3_init(void)
{

	platform_device_register(&smc91x_device);
	platform_device_register(&xyz3_audio_device);
	platform_device_register(&xyz3ostimer_device);
	platform_device_register(&xyz3adc_device);
	platform_device_register(&xyz3pwm_device);
	platform_device_register(&xyz3gpio_device);
	platform_device_register(&xyz3doc_device);
}


static struct map_desc xyz3_io_desc[] __initdata = {
  {   /* CPLD */
        .virtual    =  XYZ3_FPGA_VIRT,
        .pfn        = __phys_to_pfn(XYZ3_FPGA_PHYS),
        .length     = 0x00100000,
        .type       = MT_DEVICE
  },
};

static void __init xyz3_map_io(void)
{
	pxa_map_io();
	iotable_init(xyz3_io_desc, ARRAY_SIZE(xyz3_io_desc));

	/* initialize sleep mode regs (wake-up sources, etc) */
	PGSR0 = 0x00000000;
	PGSR1 = 0x00000000;
	PGSR2 = 0x00000000;
	PGSR3 = 0x00000000;
	PWER  = 0x00000003; // GPIO 0,1 wakeup; POWER BTN and ACC_ON
	PRER  = 0x00000000; // ditto (disable raising edge detect)
	PFER  = 0x00000003; // ditto (enable falling edge detect)
	PSLR |= 0x00400000; // Ignore VDD Fault in deep sleep mode.
}

MACHINE_START(MAINSTONE, "Sony PXA270 Development Platform (aka XYZ3)")
	.phys_ram		= 0xa0000000,
	.phys_io		= 0x40000000,
	.io_pg_offst	= (io_p2v(0x40000000) >> 18) & 0xfffc,
	.boot_params	= 0xa0000100,
	.map_io			= xyz3_map_io,
	.init_irq		= xyz3_init_irq,
	.timer			= &pxa_timer,
	.init_machine	= xyz3_init,
MACHINE_END
