
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/sched.h>
// #include <linux/malloc.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/cdev.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>

#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>

#include <asm/mach/irq.h>

#include <asm/arch/system.h>
#include <asm/arch/pxa-regs.h>
#include <asm/arch/xyz3.h>
#include <asm/arch/xyz3-devnum.h>

#include <asm/arch/xyz3gpio.h>

/* define */
#define GPIO_BASE_ADDR		0x40E00000	// 
#define GPIO_SIZE			0x200	// 

#define IRQ_NAME_LEN	32
#define XYZ3_GPIRQ_MAX	5

#define EVENT_SET_FLAG	1
#define EVENT_CLR_FLAG	0

/* Global */
typedef struct {
	unsigned int		irq_no;
	unsigned long		port;
	int					trigger;
	int					wait_flag;
	wait_queue_head_t	*wait_q;
	char				name[IRQ_NAME_LEN+1];
} xyz3_irq_table;

static xyz3_irq_table *pirq_tbl[XYZ3_GPIRQ_MAX];

DECLARE_MUTEX( InitSem );


/* Suspend/Resume  */
static struct {
	u32		PinLevel0;
	u32		PinLevel1;
	u32		PinLevel2;
	u32		PinLevel3;
} SaveReg;

//#define DEBUG
#ifdef DEBUG
#define DPRN(x)		x
#else
#define DPRN(x)
#endif


/* Device file */
#if 0
static int DevMajorNumber = 0;		// major number
#else
static int DevMajorNumber = XYZ3_MAJOR_GPIO;	// major number
#endif
static int DevMinorNumber = XYZ3_MINOR_GPIO;	// minor number
static int DevCount = 1;			// 
static char *DevName = "gpio";		// device driver name


// Open() counter.
static int Open_count = 0;

int xyz3GpioGetLevel(unsigned long arg)
{
	unsigned long bit;
	int stat = 0;

	DPRN( printk( KERN_DEBUG "%s(0x%08lx) called.\n", __FUNCTION__, arg ) );

	if( ( arg & XYZ3_GP_HIGH_WORD ) == XYZ3_GP_HIGH_WORD )
	{
		bit = (arg & XYZ3_GPIO_MASK) << 16;
	}
	else
	{
		bit = arg & XYZ3_GPIO_MASK;
	}

	DPRN( printk( KERN_DEBUG "%s:bit=%08lx\n", __FUNCTION__, bit ) );
	if( ( arg & XYZ3_GPREG_MASK ) == XYZ3_GPxR0 )
	{
		DPRN( printk( KERN_DEBUG "%s:GPLR0\n", __FUNCTION__ ) );
		if( ( GPLR0 & bit ) == bit )
		{
			stat = 1;
		}
	}
	else if( ( arg & XYZ3_GPREG_MASK ) == XYZ3_GPxR1 )
	{
		DPRN( printk( KERN_DEBUG "%s:GPLR3\n", __FUNCTION__ ) );
		if( ( GPLR1 & bit ) == bit )
		{
			stat = 1;
		}
	}
	else  if( ( arg & XYZ3_GPREG_MASK ) == XYZ3_GPxR2 )
	{
		DPRN( printk( KERN_DEBUG "%s:GPLR3\n", __FUNCTION__ ) );
		if( ( GPLR2 & bit ) == bit )
		{
			stat = 1;
		}
	}
	else  if( ( arg & XYZ3_GPREG_MASK ) == XYZ3_GPxR3 )
	{
		DPRN( printk( KERN_DEBUG "%s:GPLR3\n", __FUNCTION__ ) );
		if( ( GPLR3 & bit ) == bit )
		{
			stat = 1;
		}
	}
	else
	{
		stat = -EINVAL;
	}

	return stat;
}

int xyz3GpioSetLevel(unsigned long arg, int level)
{
	unsigned long bit;
	int stat = 0;

	DPRN( printk( KERN_DEBUG "%s(0x%08lx, %x) called.\n", __FUNCTION__, arg, level) );

	if( ( arg & XYZ3_GP_HIGH_WORD ) == XYZ3_GP_HIGH_WORD )
	{
		bit = (arg & XYZ3_GPIO_MASK) << 16;
	}
	else
	{
		bit = arg & XYZ3_GPIO_MASK;
	}

	DPRN( printk( KERN_DEBUG "%s:bit=%08lx\n", __FUNCTION__, bit ) );
	if( level == 0 )
	{
		DPRN( printk( KERN_DEBUG "%s:Low\n", __FUNCTION__ ) );
		if( ( arg & XYZ3_GPREG_MASK ) == XYZ3_GPxR0 )
		{
			DPRN( printk( KERN_DEBUG "%s:GPCR0\n", __FUNCTION__ ) );
			GPCR0 |=  bit;
		}
		else if( ( arg & XYZ3_GPREG_MASK ) == XYZ3_GPxR1 )
		{
			DPRN( printk( KERN_DEBUG "%s:GPCR1\n", __FUNCTION__ ) );
			GPCR1 |=  bit;
		}
		else  if( ( arg & XYZ3_GPREG_MASK ) == XYZ3_GPxR2 )
		{
			DPRN( printk( KERN_DEBUG "%s:GPCR2\n", __FUNCTION__ ) );
			GPCR2 |=  bit;
		}
		else  if( ( arg & XYZ3_GPREG_MASK ) == XYZ3_GPxR3 )
		{
			DPRN( printk( KERN_DEBUG "%s:GPCR3\n", __FUNCTION__ ) );
			GPCR3 |=  bit;
		}
		else
		{
			stat = -EINVAL;
		}
	}
	else
	{
		DPRN( printk( KERN_DEBUG "%s:High\n", __FUNCTION__ ) );
		if( ( arg & XYZ3_GPREG_MASK ) == XYZ3_GPxR0 )
		{
			DPRN( printk( KERN_DEBUG "%s:GPSR0\n", __FUNCTION__ ) );
			GPSR0 |=  bit;
		}
		else if( ( arg & XYZ3_GPREG_MASK ) == XYZ3_GPxR1 )
		{
			DPRN( printk( KERN_DEBUG "%s:GPSR1\n", __FUNCTION__ ) );
			GPSR1 |=  bit;
		}
		else  if( ( arg & XYZ3_GPREG_MASK ) == XYZ3_GPxR2 )
		{
			DPRN( printk( KERN_DEBUG "%s:GPSR2\n", __FUNCTION__ ) );
			GPSR2 |=  bit;
		}
		else  if( ( arg & XYZ3_GPREG_MASK ) == XYZ3_GPxR3 )
		{
			DPRN( printk( KERN_DEBUG "%s:GPSR3\n", __FUNCTION__ ) );
			GPSR3 |=  bit;
		}
		else
		{
			stat = -EINVAL;
		}
	}

	return stat;
}


static irqreturn_t gpio_int( int irq_no, void *dev_id, struct pt_regs *regs )
{
	int	i;
	int	find = 0;

	DPRN( printk( KERN_DEBUG "%s:called. irq_no = %d\n", __FUNCTION__, irq_no) );
	for( i=0; i< XYZ3_GPIRQ_MAX; i++ )
	{
		if( pirq_tbl[i] != NULL )
		{
			if( pirq_tbl[i]->irq_no == irq_no )
			{
				find = 1;
				break;
			}
		}
	}

	if( find == 0 )
	{
		printk( KERN_ERR "%s:Irq not find!\n", __FUNCTION__ );
		return IRQ_HANDLED;
	}

	pirq_tbl[i]->wait_flag = EVENT_SET_FLAG;
	wake_up_interruptible( pirq_tbl[i]->wait_q );
	return IRQ_HANDLED;
}

int xyz3GpioIrqInit( Xyz3IrqInitArg *arg )
{
	int	namelen;
	int	i;
	int	first_entry = -1;
	int	irq_no;
	int	edgetype;
	int stat = 0;
	irq_no = (arg->port & XYZ3_GPNUM_MASK) >> 24;
	irq_no = IRQ_GPIO(irq_no);

	DPRN( printk( KERN_DEBUG "%s:called. irq_no = %d\n", __FUNCTION__, irq_no) );

	for( i=0; i< XYZ3_GPIRQ_MAX; i++ )
	{
		//DPRN( printk( KERN_DEBUG "xyz3GpioIrqInit:pirq_tbl[%d] = 0x%lx\n", i, pirq_tbl[i] ) );
		if( pirq_tbl[i] != NULL )
		{
			//DPRN( printk( KERN_DEBUG "xyz3GpioIrqInit:pirq_tbl[%d]->irq_no = %d\n", i, pirq_tbl[i]->irq_no) );
			if( pirq_tbl[i]->irq_no == irq_no )
			{
				printk( KERN_ERR "%s:It already exists.\n", __FUNCTION__ );
				return -EPERM;
			}
		}
		else
		{
			if( first_entry == -1 )
			{
				first_entry = i;
			}
		}
	}

	down( &InitSem );
	if( first_entry != -1 )
	{
		//DPRN( printk( KERN_DEBUG "%s:Entry = %d\n", __FUNCTION__, first_entry) );
		//DPRN( printk( KERN_DEBUG "%s:sizeof(xyz3_irq_table):%d\n", __FUNCTION__, sizeof(xyz3_irq_table) ) );

		pirq_tbl[first_entry] = kmalloc(sizeof(xyz3_irq_table), GFP_KERNEL);
		if( pirq_tbl[first_entry] == NULL )
		{
			printk( KERN_ERR "%s:kmalloc(xyz3_irq_table) error!\n", __FUNCTION__ );
			stat = -ENOMEM;
			goto InitErr;
		}

		pirq_tbl[first_entry]->wait_q   = kmalloc(sizeof(wait_queue_head_t), GFP_KERNEL);
		if( pirq_tbl[first_entry]->wait_q == NULL )
		{
			printk( KERN_ERR "%s:kmalloc(wait_queue_t) error!\n", __FUNCTION__ );
			pirq_tbl[first_entry] = NULL;
			stat = -ENOMEM;
			goto InitErr;
		}

		init_waitqueue_head( pirq_tbl[first_entry]->wait_q );
		pirq_tbl[first_entry]->irq_no    = irq_no;
		pirq_tbl[first_entry]->port      = arg->port;
		pirq_tbl[first_entry]->trigger   = arg->trigger;
		pirq_tbl[first_entry]->wait_flag = EVENT_CLR_FLAG;

		memset( pirq_tbl[first_entry]->name, 0, IRQ_NAME_LEN + 1 );
		namelen = strlen( arg->name );
		if( namelen > IRQ_NAME_LEN )
		{
			memcpy( pirq_tbl[first_entry]->name, arg->name, IRQ_NAME_LEN );
		}
		else
		{
			memcpy( pirq_tbl[first_entry]->name, arg->name, namelen );
		}

		switch( arg->trigger )
		{
			case XYZ3_IRQ_FALL_EDGE:   edgetype = IRQT_FALLING; break;
			case XYZ3_IRQ_RISE_EDGE:   edgetype = IRQT_RISING;  break;
			case XYZ3_IRQ_DOUBLE_EDGE: edgetype = (IRQT_FALLING | IRQT_RISING); break;
			case XYZ3_IRQ_LOW_LEVEL:   edgetype = IRQT_FALLING; break;
			case XYZ3_IRQ_HIGH_LEVEL:  edgetype = IRQT_RISING;  break;
			default:
				printk( KERN_ERR "%s:Invalid triger type!!\n", __FUNCTION__ );
				stat = -EINVAL;
				goto InitErr;
		}

		set_irq_type( irq_no, edgetype );
		request_irq( irq_no, gpio_int, SA_INTERRUPT, pirq_tbl[first_entry]->name, NULL );
	}
	else
	{
		printk( KERN_ERR "%s:Irq Table is full!\n", __FUNCTION__ );
	}

	up( &InitSem );

	return 0;

InitErr:
	if( first_entry != -1 )
	{
		if( pirq_tbl[first_entry] != NULL )
		{
			if( pirq_tbl[first_entry]->wait_q != NULL )
			{
				kfree( pirq_tbl[first_entry]->wait_q );
				pirq_tbl[first_entry]->wait_q = NULL;
			}
			kfree( pirq_tbl[first_entry] );
			pirq_tbl[first_entry] = NULL;
		}
	}

	up( &InitSem );
	return stat;
}

int xyz3GpioIrqEnable( unsigned long port )
{
	int	i;
	int	find = 0;
	int irq_no;
	irq_no = (port & XYZ3_GPNUM_MASK) >> 24;
	irq_no = IRQ_GPIO(irq_no);

	DPRN( printk( KERN_DEBUG "%s:called. irq_no = %d\n", __FUNCTION__, irq_no) );

	for( i=0; i< XYZ3_GPIRQ_MAX; i++ )
	{
		if( pirq_tbl[i] != NULL )
		{
			if( pirq_tbl[i]->irq_no == irq_no )
			{
				find = 1;
				break;
			}
		}
	}

	if( find == 0 )
	{
		printk( KERN_ERR "%s:Port not find!\n", __FUNCTION__ );
		return -EINVAL;
	}

	enable_irq( irq_no );
	return 0;
}

int xyz3GpioIrqDisable( unsigned long port )
{
	int	i;
	int	find = 0;
	int irq_no;
	irq_no = (port & XYZ3_GPNUM_MASK) >> 24;
	irq_no = IRQ_GPIO(irq_no);

	DPRN( printk( KERN_DEBUG "%s:called. irq_no = %d\n", __FUNCTION__, irq_no) );

	for( i=0; i< XYZ3_GPIRQ_MAX; i++ )
	{
		if( pirq_tbl[i] != NULL )
		{
			if( pirq_tbl[i]->irq_no == irq_no )
			{
				find = 1;
				break;
			}
		}
	}

	if( find == 0 )
	{
		printk( KERN_ERR "%s:Port not find!\n", __FUNCTION__ );
		return -EINVAL;
	}

	disable_irq( irq_no );
	return 0;
}

int xyz3GpioIrqWait( unsigned long port )
{
	int	i;
	int	find = 0;
	int	irq_no;
	int	rc;
	int portlevel;
	irq_no = (port & XYZ3_GPNUM_MASK) >> 24;
	irq_no = IRQ_GPIO(irq_no);

	DPRN( printk( KERN_DEBUG "%s:called. irq_no = %d\n", __FUNCTION__, irq_no) );

	for( i=0; i< XYZ3_GPIRQ_MAX; i++ )
	{
		if( pirq_tbl[i] != NULL )
		{
			if( pirq_tbl[i]->irq_no == irq_no )
			{
				find = 1;
				break;
			}
		}
	}

	if( find == 0 )
	{
		printk( KERN_ERR "%s:Port not find!\n", __FUNCTION__ );
		return -EINVAL;
	}

	// for level trigger
	portlevel = xyz3GpioGetLevel( port );
	if( ( (pirq_tbl[i]->trigger == XYZ3_IRQ_LOW_LEVEL)  && (portlevel == 0) )
	 || ( (pirq_tbl[i]->trigger == XYZ3_IRQ_HIGH_LEVEL) && (portlevel == 1) ) )
	{
		return portlevel;
	}

	rc = wait_event_interruptible( *(pirq_tbl[i]->wait_q), (pirq_tbl[i]->wait_flag != EVENT_CLR_FLAG) );
	pirq_tbl[i]->wait_flag = EVENT_CLR_FLAG;

	if( rc != 0 )
	{
		//return rc;		// -ERESTARTSYS (-200)
		return -EAGAIN;
	}
	return xyz3GpioGetLevel( port );
}

int xyz3GpioIrqClean( unsigned long port )
{
	int	i;
	int	find = 0;
	int irq_no;
	irq_no = (port & XYZ3_GPNUM_MASK) >> 24;
	irq_no = IRQ_GPIO(irq_no);

	DPRN( printk( KERN_DEBUG "%s:called. irq_no = %d\n", __FUNCTION__, irq_no) );

	for( i=0; i< XYZ3_GPIRQ_MAX; i++ )
	{
		//DPRN( printk( KERN_DEBUG "%s:pirq_tbl[%d] = 0x%lx\n", __FUNCTION__, i, pirq_tbl[i]) );
		if( pirq_tbl[i] != NULL )
		{
			if( pirq_tbl[i]->irq_no == irq_no )
			{
				find = 1;
				break;
			}
		}
	}

	if( find == 0 )
	{
		printk( KERN_ERR "%s:Port not find!\n", __FUNCTION__ );
		return -EINVAL;
	}

	free_irq( irq_no, NULL );

	down( &InitSem );
	kfree( pirq_tbl[i]->wait_q);
	kfree( pirq_tbl[i] );
	pirq_tbl[i] = NULL;
	up( &InitSem );

	return 0;
}

int xyz3GpioIrqStop( unsigned long port )
{
	int	i;
	int	find = 0;
	int irq_no;
	irq_no = (port & XYZ3_GPNUM_MASK) >> 24;
	irq_no = IRQ_GPIO(irq_no);

	DPRN( printk( KERN_DEBUG "%s:called. irq_no = %d\n", __FUNCTION__, irq_no) );

	for( i=0; i< XYZ3_GPIRQ_MAX; i++ )
	{
		if( pirq_tbl[i] != NULL )
		{
			if( pirq_tbl[i]->irq_no == irq_no )
			{
				find = 1;
				break;
			}
		}
	}

	if( find == 0 )
	{
		printk( KERN_ERR "%s:Port not find!\n", __FUNCTION__ );
		return -EINVAL;
	}

	pirq_tbl[i]->wait_flag = EVENT_SET_FLAG;
	wake_up_interruptible( pirq_tbl[i]->wait_q );

	return 0;
}

static void irq_tbl_init( void )
{
	int i;

	DPRN( printk( KERN_DEBUG "%s:called.\n", __FUNCTION__) );

	for( i=0; i< XYZ3_GPIRQ_MAX; i++ )
	{
		pirq_tbl[i] = NULL;
	}

}

static void irq_tbl_clean( void )
{
	int	i;

	DPRN( printk( KERN_DEBUG "%s:called.\n", __FUNCTION__ ) );

	for( i=0; i< XYZ3_GPIRQ_MAX; i++ )
	{
		if( pirq_tbl[i] != NULL )
		{
			if( pirq_tbl[i]->wait_q != NULL )
			{
				kfree( pirq_tbl[i]->wait_q);
			}
			free_irq( pirq_tbl[i]->irq_no, NULL );
			kfree( pirq_tbl[i] );
			pirq_tbl[i] = NULL;
		}
	}
}


static int gpio_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	int stat = 0;

	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );

	switch( cmd )
	{
		case XYZ3GP_GET_LEVEL:
			stat = xyz3GpioGetLevel(arg);
			break;
		case XYZ3GP_SET_HIGH:
			stat = xyz3GpioSetLevel(arg, 1);
			break;
		case XYZ3GP_SET_LOW:
			stat = xyz3GpioSetLevel(arg, 0);
			break;

		case XYZ3GP_IRQ_INIT:
			stat = xyz3GpioIrqInit( (Xyz3IrqInitArg*)arg );
			break;

		case XYZ3GP_IRQ_ENABLE:
			stat = xyz3GpioIrqEnable( arg );
			break;

		case XYZ3GP_IRQ_DISABLE:
			stat = xyz3GpioIrqDisable( arg );
			break;

		case XYZ3GP_IRQ_WAIT:
			stat = xyz3GpioIrqWait( arg );
			break;

		case XYZ3GP_IRQ_STOP:
			stat = xyz3GpioIrqStop( arg );
			break;

		case XYZ3GP_IRQ_CLEAN:
			stat = xyz3GpioIrqClean( arg );
			break;

		default:
			stat = -EINVAL;
			break;
	}
	return stat;
}


static int gpio_open(struct inode *inode, struct file *filp)
{
	Open_count++;
	return 0;
}


static int gpio_release( struct inode *inode, struct file *filp )
{
	if( Open_count < 0 )
	{
		return -EPERM;
	}
	Open_count--; 
	return 0;
}


static struct file_operations gpio_fops = {
	owner:	THIS_MODULE,
	ioctl:	gpio_ioctl,
	open:	gpio_open,
	release: gpio_release,
};


/*
 *  Power manager Support Routines
 */
static int __init gpio_probe(struct platform_device *dev)
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );
	memset( &SaveReg, 0, sizeof( SaveReg ) );
	return 0;
}

static int gpio_remove(struct platform_device *dev)
{
	DPRN( printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ ) );
	return 0;
}

static int gpio_suspend(struct platform_device *dev, pm_message_t state)
{
	printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ );

	SaveReg.PinLevel0 = GPLR0;
	SaveReg.PinLevel1 = GPLR1;
	SaveReg.PinLevel2 = GPLR2;
	SaveReg.PinLevel3 = GPLR3;

	return 0;
}

static int gpio_resume(struct platform_device *dev)
{
	printk( KERN_DEBUG "%s() called.\n", __FUNCTION__ );

	GPSR0 =  SaveReg.PinLevel0;
	GPCR0 = ~SaveReg.PinLevel0;
	GPSR1 =  SaveReg.PinLevel1;
	GPCR1 = ~SaveReg.PinLevel1;
	GPSR2 =  SaveReg.PinLevel2;
	GPCR2 = ~SaveReg.PinLevel2;
	GPSR3 =  SaveReg.PinLevel3;
	GPCR3 = ~SaveReg.PinLevel3;

	return 0;
}

static struct platform_driver xyz3gpio_driver = {
	.probe		= gpio_probe,
	.remove		= gpio_remove,
	.suspend	= gpio_suspend,
	.resume		= gpio_resume,
	.driver		= {
		.name	= "xyz3-gpio",
	},
};



static struct cdev gpio_cdev;

static int __init gpio_init( void )
{
	int result;
	int err;
	int devno;
	dev_t dev = 0;

	// Actually allocate io-port region. Always success.
	if( request_mem_region(GPIO_BASE_ADDR, GPIO_SIZE, DevName) == NULL )
	{
		printk( KERN_ERR "%s:request_mem_region() error!!\n", __FUNCTION__ );
		return -EBUSY;
	}

	// Append this device driver to system. 
	if (DevMajorNumber)
	{
		dev = MKDEV(DevMajorNumber, DevMinorNumber);
		result = register_chrdev_region(dev, DevCount, DevName);
	}
	else
	{
		result = alloc_chrdev_region(&dev, DevMinorNumber, DevCount, DevName);
		DevMajorNumber = MAJOR(dev);
	}
	if (result < 0) 
	{
		printk( KERN_ERR "%s:can't get major %d\n", __FUNCTION__, DevMinorNumber);
		return result;
	}

	devno = MKDEV(DevMajorNumber, DevMinorNumber);
	cdev_init(&gpio_cdev, &gpio_fops);
	gpio_cdev.owner = THIS_MODULE;
	gpio_cdev.ops = &gpio_fops;

	err = cdev_add(&gpio_cdev, devno, 1);
	if( err )
	{
		printk( KERN_ERR "%s:Error %d adding msys_dev!!\n", __FUNCTION__, err);
	}

	irq_tbl_init();

	result = platform_driver_register(&xyz3gpio_driver);
	if( result != 0 )
	{
		printk(KERN_ERR "%s:driver_register() failed. (%d)", __FUNCTION__, result);
	}

	return result;
}


static void __exit gpio_cleanup(void)
{
	dev_t devno = MKDEV(DevMajorNumber, DevMinorNumber);

	platform_driver_unregister(&xyz3gpio_driver);

	irq_tbl_clean();

	cdev_del(&gpio_cdev);

	release_mem_region(GPIO_BASE_ADDR, GPIO_SIZE);
	unregister_chrdev_region(devno, DevCount);
}


module_init(gpio_init);
module_exit(gpio_cleanup);

EXPORT_SYMBOL(xyz3GpioGetLevel);
EXPORT_SYMBOL(xyz3GpioSetLevel);
EXPORT_SYMBOL(xyz3GpioIrqInit);
EXPORT_SYMBOL(xyz3GpioIrqEnable);
EXPORT_SYMBOL(xyz3GpioIrqDisable);
EXPORT_SYMBOL(xyz3GpioIrqWait);
EXPORT_SYMBOL(xyz3GpioIrqClean);
EXPORT_SYMBOL(xyz3GpioIrqStop);

MODULE_LICENSE("GPL");

